/*
 * kgraft_patch_bsc1008284
 *
 * Fix for bsc#1008284
 *
 *  Upstream commit:
 *  a9b1fc851db0 ("Btrfs: fix endless loop in balancing block groups")
 *
 *  SLE12-SP2 commit:
 *  2659fe9d134c63d2289572247e5f2e1caff03602
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/unaligned/access_ok.h>

#include "local.h"

#define KGR_PATCHED_MODULE      "btrfs"

static struct btrfs_path *(*kgr_btrfs_alloc_path)(void);
static int (*kgr_prepare_to_relocate)(struct reloc_control *rc);
static int (*kgr_btrfs_block_rsv_refill)(struct btrfs_root *root, struct btrfs_block_rsv *block_rsv, u64 min_reserved, enum btrfs_reserve_flush_enum flush);
static struct btrfs_trans_handle *(*kgr_btrfs_start_transaction)(struct btrfs_root *root, unsigned int num_items);
static int (*kgr_btrfs_end_transaction)(struct btrfs_trans_handle *trans, struct btrfs_root *root);
static void (*kgr_remove_backref_node)(struct backref_cache *cache, struct backref_node *node);
static void (*kgr_update_backref_node)(struct backref_cache *cache, struct backref_node *node, u64 bytenr);
static int (*kgr_find_next_extent)(struct reloc_control *rc, struct btrfs_path *path, struct btrfs_key *extent_key);
static int (*kgr_btrfs_next_leaf)(struct btrfs_root *root, struct btrfs_path *path);
static void (*kgr_read_extent_buffer)(struct extent_buffer *eb, void *dstv, unsigned long start, unsigned long len);
static void (*kgr_btrfs_release_path)(struct btrfs_path *p);
static int (*kgr_btrfs_search_slot)(struct btrfs_trans_handle *trans, struct btrfs_root *root, struct btrfs_key *key, struct btrfs_path *p, int ins_len, int cow);
static int (*kgr_add_tree_block)(struct reloc_control *rc, struct btrfs_key *extent_key, struct btrfs_path *path, struct rb_root *blocks);
static int (*kgr_add_data_references)(struct reloc_control *rc, struct btrfs_key *extent_key, struct btrfs_path *path, struct rb_root *blocks);
static int (*kgr_relocate_tree_blocks)(struct btrfs_trans_handle *trans, struct reloc_control *rc, struct rb_root *blocks);
static int (*kgr_btrfs_end_transaction_throttle)(struct btrfs_trans_handle *trans, struct btrfs_root *root);
static void (*kgr_btrfs_btree_balance_dirty)(struct btrfs_root *root);
static int (*kgr_relocate_data_extent)(struct inode *inode, struct btrfs_key *extent_key, struct file_extent_cluster *cluster);
static int (*kgr_btrfs_force_chunk_alloc)(struct btrfs_trans_handle *trans, struct btrfs_root *root, u64 type);
static int (*kgr_clear_extent_bits)(struct extent_io_tree *tree, u64 start, u64 end, unsigned bits, gfp_t mask);
static int (*kgr_relocate_file_extent_cluster)(struct inode *inode, struct file_extent_cluster *cluster);
static void (*kgr_set_reloc_control)(struct reloc_control *rc);
static void (*kgr_btrfs_block_rsv_release)(struct btrfs_root *root, struct btrfs_block_rsv *block_rsv, u64 num_bytes);
static int (*kgr_prepare_to_merge)(struct reloc_control *rc, int err);
static void (*kgr_merge_reloc_roots)(struct reloc_control *rc);
static struct btrfs_trans_handle *(*kgr_btrfs_join_transaction)(struct btrfs_root *root);
static int (*kgr_lock_extent)(struct extent_io_tree *tree, u64 start, u64 end);
static int (*kgr_btrfs_qgroup_insert_dirty_extent)(struct btrfs_trans_handle *trans, struct btrfs_fs_info *fs_info, u64 bytenr, u64 num_bytes, gfp_t gfp_flag);
static int (*kgr_unlock_extent)(struct extent_io_tree *tree, u64 start, u64 end);
static void (*kgr_btrfs_free_path)(struct btrfs_path *p);
static int (*kgr_btrfs_next_old_leaf)(struct btrfs_root *root, struct btrfs_path *path, u64 time_seq);
static void (*kgr__btrfs_abort_transaction)(struct btrfs_trans_handle *trans, struct btrfs_root *root, const char *function, unsigned int line, int errno);
static int (*kgr_btrfs_commit_transaction)(struct btrfs_trans_handle *trans, struct btrfs_root *root);
static void (*kgr_btrfs_free_block_rsv)(struct btrfs_root *root, struct btrfs_block_rsv *rsv);
static int (*kgr_map_private_extent_buffer)(struct extent_buffer *eb, unsigned long start, unsigned long min_len, char **map, unsigned long *map_start, unsigned long *map_len);
static void (*kgr_write_extent_buffer)(struct extent_buffer *eb, const void *srcv, unsigned long start, unsigned long len);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "btrfs_alloc_path", (void *)&kgr_btrfs_alloc_path },
	{ "prepare_to_relocate", (void *)&kgr_prepare_to_relocate },
	{ "btrfs_block_rsv_refill", (void *)&kgr_btrfs_block_rsv_refill },
	{ "btrfs_start_transaction", (void *)&kgr_btrfs_start_transaction },
	{ "btrfs_end_transaction", (void *)&kgr_btrfs_end_transaction },
	{ "remove_backref_node", (void *)&kgr_remove_backref_node },
	{ "update_backref_node", (void *)&kgr_update_backref_node },
	{ "find_next_extent", (void *)&kgr_find_next_extent },
	{ "btrfs_next_leaf", (void *)&kgr_btrfs_next_leaf },
	{ "read_extent_buffer", (void *)&kgr_read_extent_buffer },
	{ "btrfs_release_path", (void *)&kgr_btrfs_release_path },
	{ "btrfs_search_slot", (void *)&kgr_btrfs_search_slot },
	{ "add_tree_block", (void *)&kgr_add_tree_block },
	{ "add_data_references", (void *)&kgr_add_data_references },
	{ "relocate_tree_blocks", (void *)&kgr_relocate_tree_blocks },
	{ "btrfs_end_transaction_throttle", (void *)&kgr_btrfs_end_transaction_throttle },
	{ "btrfs_btree_balance_dirty", (void *)&kgr_btrfs_btree_balance_dirty },
	{ "relocate_data_extent", (void *)&kgr_relocate_data_extent },
	{ "btrfs_force_chunk_alloc", (void *)&kgr_btrfs_force_chunk_alloc },
	{ "clear_extent_bits", (void *)&kgr_clear_extent_bits },
	{ "relocate_file_extent_cluster", (void *)&kgr_relocate_file_extent_cluster },
	{ "set_reloc_control", (void *)&kgr_set_reloc_control },
	{ "btrfs_block_rsv_release", (void *)&kgr_btrfs_block_rsv_release },
	{ "prepare_to_merge", (void *)&kgr_prepare_to_merge },
	{ "merge_reloc_roots", (void *)&kgr_merge_reloc_roots },
	{ "btrfs_join_transaction", (void *)&kgr_btrfs_join_transaction },
	{ "lock_extent", (void *)&kgr_lock_extent },
	{ "btrfs_qgroup_insert_dirty_extent", (void *)&kgr_btrfs_qgroup_insert_dirty_extent },
	{ "unlock_extent", (void *)&kgr_unlock_extent },
	{ "btrfs_free_path", (void *)&kgr_btrfs_free_path },
	{ "btrfs_next_old_leaf", (void *)&kgr_btrfs_next_old_leaf },
	{ "__btrfs_abort_transaction", (void *)&kgr__btrfs_abort_transaction },
	{ "btrfs_commit_transaction", (void *)&kgr_btrfs_commit_transaction },
	{ "btrfs_free_block_rsv", (void *)&kgr_btrfs_free_block_rsv },
	{ "map_private_extent_buffer", (void *)&kgr_map_private_extent_buffer },
	{ "write_extent_buffer", (void *)&kgr_write_extent_buffer },
};

#define DEFINE_BTRFS_SETGET_BITS(bits)					\
u##bits btrfs_get_token_##bits(struct extent_buffer *eb, void *ptr,	\
			       unsigned long off,			\
			       struct btrfs_map_token *token)		\
{									\
	unsigned long part_offset = (unsigned long)ptr;			\
	unsigned long offset = part_offset + off;			\
	void *p;							\
	int err;							\
	char *kaddr;							\
	unsigned long map_start;					\
	unsigned long map_len;						\
	int size = sizeof(u##bits);					\
	u##bits res;							\
									\
	if (token && token->kaddr && token->offset <= offset &&		\
	    token->eb == eb &&						\
	   (token->offset + PAGE_CACHE_SIZE >= offset + size)) {	\
		kaddr = token->kaddr;					\
		p = kaddr + part_offset - token->offset;		\
		res = get_unaligned_le##bits(p + off);			\
		return res;						\
	}								\
	err = kgr_map_private_extent_buffer(eb, offset, size,		\
					&kaddr, &map_start, &map_len);	\
	if (err) {							\
		__le##bits leres;					\
									\
		kgr_read_extent_buffer(eb, &leres, offset, size);		\
		return le##bits##_to_cpu(leres);			\
	}								\
	p = kaddr + part_offset - map_start;				\
	res = get_unaligned_le##bits(p + off);				\
	if (token) {							\
		token->kaddr = kaddr;					\
		token->offset = map_start;				\
		token->eb = eb;						\
	}								\
	return res;							\
}									\
void btrfs_set_token_##bits(struct extent_buffer *eb,			\
			    void *ptr, unsigned long off, u##bits val,	\
			    struct btrfs_map_token *token)		\
{									\
	unsigned long part_offset = (unsigned long)ptr;			\
	unsigned long offset = part_offset + off;			\
	void *p;							\
	int err;							\
	char *kaddr;							\
	unsigned long map_start;					\
	unsigned long map_len;						\
	int size = sizeof(u##bits);					\
									\
	if (token && token->kaddr && token->offset <= offset &&		\
	    token->eb == eb &&						\
	   (token->offset + PAGE_CACHE_SIZE >= offset + size)) {	\
		kaddr = token->kaddr;					\
		p = kaddr + part_offset - token->offset;		\
		put_unaligned_le##bits(val, p + off);			\
		return;							\
	}								\
	err = kgr_map_private_extent_buffer(eb, offset, size,		\
			&kaddr, &map_start, &map_len);			\
	if (err) {							\
		__le##bits val2;					\
									\
		val2 = cpu_to_le##bits(val);				\
		kgr_write_extent_buffer(eb, &val2, offset, size);		\
		return;							\
	}								\
	p = kaddr + part_offset - map_start;				\
	put_unaligned_le##bits(val, p + off);				\
	if (token) {							\
		token->kaddr = kaddr;					\
		token->offset = map_start;				\
		token->eb = eb;						\
	}								\
}

DEFINE_BTRFS_SETGET_BITS(8)
DEFINE_BTRFS_SETGET_BITS(16)
DEFINE_BTRFS_SETGET_BITS(32)
DEFINE_BTRFS_SETGET_BITS(64)

/* optimized function */
static int kgr_update_backref_cache(struct btrfs_trans_handle *trans,
				struct backref_cache *cache)
{
	struct backref_node *node;
	int level = 0;

	if (cache->last_trans == 0) {
		cache->last_trans = trans->transid;
		return 0;
	}

	if (cache->last_trans == trans->transid)
		return 0;

	/*
	 * detached nodes are used to avoid unnecessary backref
	 * lookup. transaction commit changes the extent tree.
	 * so the detached nodes are no longer useful.
	 */
	while (!list_empty(&cache->detached)) {
		node = list_entry(cache->detached.next,
				  struct backref_node, list);
		kgr_remove_backref_node(cache, node);
	}

	while (!list_empty(&cache->changed)) {
		node = list_entry(cache->changed.next,
				  struct backref_node, list);
		list_del_init(&node->list);
		BUG_ON(node->pending);
		kgr_update_backref_node(cache, node, node->new_bytenr);
	}

	/*
	 * some nodes can be left in the pending list if there were
	 * errors during processing the pending nodes.
	 */
	for (level = 0; level < BTRFS_MAX_LEVEL; level++) {
		list_for_each_entry(node, &cache->pending[level], list) {
			BUG_ON(!node->pending);
			if (node->bytenr == node->new_bytenr)
				continue;
			kgr_update_backref_node(cache, node, node->new_bytenr);
		}
	}

	cache->last_trans = 0;
	return 1;
}

/* inlined */
static int kgr_check_extent_flags(u64 flags)
{
	if ((flags & BTRFS_EXTENT_FLAG_DATA) &&
	    (flags & BTRFS_EXTENT_FLAG_TREE_BLOCK))
		return 1;
	if (!(flags & BTRFS_EXTENT_FLAG_DATA) &&
	    !(flags & BTRFS_EXTENT_FLAG_TREE_BLOCK))
		return 1;
	if ((flags & BTRFS_EXTENT_FLAG_DATA) &&
	    (flags & BTRFS_BLOCK_FLAG_FULL_BACKREF))
		return 1;
	return 0;
}

/* inlined */
static inline void kgr_btrfs_item_key(struct extent_buffer *eb,
			   struct btrfs_disk_key *disk_key, int nr)
{
	struct btrfs_item *item = btrfs_item_nr(nr);
	read_eb_member(eb, item, struct btrfs_item, key, disk_key);
}

/* inlined */
static inline void kgr_btrfs_disk_key_to_cpu(struct btrfs_key *cpu,
					 struct btrfs_disk_key *disk)
{
	cpu->offset = le64_to_cpu(disk->offset);
	cpu->type = disk->type;
	cpu->objectid = le64_to_cpu(disk->objectid);
}

/* inlined */
static inline void kgr_btrfs_item_key_to_cpu(struct extent_buffer *eb,
				  struct btrfs_key *key, int nr)
{
	struct btrfs_disk_key disk_key;
	kgr_btrfs_item_key(eb, &disk_key, nr);
	kgr_btrfs_disk_key_to_cpu(key, &disk_key);
}

/* isra optimized */
static int kgr_get_ref_objectid_v0(struct reloc_control *rc,
			       struct btrfs_path *path,
			       struct btrfs_key *extent_key,
			       u64 *ref_objectid, int *path_change)
{
	struct btrfs_key key;
	struct extent_buffer *leaf;
	struct btrfs_extent_ref_v0 *ref0;
	int ret;
	int slot;

	leaf = path->nodes[0];
	slot = path->slots[0];
	while (1) {
		if (slot >= btrfs_header_nritems(leaf)) {
			ret = kgr_btrfs_next_leaf(rc->extent_root, path);
			if (ret < 0)
				return ret;
			BUG_ON(ret > 0);
			leaf = path->nodes[0];
			slot = path->slots[0];
			if (path_change)
				*path_change = 1;
		}
		kgr_btrfs_item_key_to_cpu(leaf, &key, slot);
		if (key.objectid != extent_key->objectid)
			return -ENOENT;

		if (key.type != BTRFS_EXTENT_REF_V0_KEY) {
			slot++;
			continue;
		}
		ref0 = btrfs_item_ptr(leaf, slot,
				struct btrfs_extent_ref_v0);
		*ref_objectid = btrfs_ref_objectid_v0(leaf, ref0);
		break;
	}
	return 0;
}

/* inlined */
static void kgr_backref_cache_cleanup(struct backref_cache *cache)
{
	struct backref_node *node;
	int i;

	while (!list_empty(&cache->detached)) {
		node = list_entry(cache->detached.next,
				  struct backref_node, list);
		kgr_remove_backref_node(cache, node);
	}

	while (!list_empty(&cache->leaves)) {
		node = list_entry(cache->leaves.next,
				  struct backref_node, lower);
		kgr_remove_backref_node(cache, node);
	}

	cache->last_trans = 0;

	for (i = 0; i < BTRFS_MAX_LEVEL; i++)
		BUG_ON(!list_empty(&cache->pending[i]));
	BUG_ON(!list_empty(&cache->changed));
	BUG_ON(!list_empty(&cache->detached));
	BUG_ON(!RB_EMPTY_ROOT(&cache->rb_root));
	BUG_ON(cache->nr_nodes);
	BUG_ON(cache->nr_edges);
}

/* isra optimized */
static void kgr_unset_reloc_control(struct reloc_control *rc)
{
	struct btrfs_fs_info *fs_info = rc->extent_root->fs_info;

	mutex_lock(&fs_info->reloc_mutex);
	fs_info->reloc_ctl = NULL;
	mutex_unlock(&fs_info->reloc_mutex);
}

/* inlined */
static inline int kgr_btrfs_next_old_item(struct btrfs_root *root,
				      struct btrfs_path *p, u64 time_seq)
{
	++p->slots[0];
	if (p->slots[0] >= btrfs_header_nritems(p->nodes[0]))
		return kgr_btrfs_next_old_leaf(root, p, time_seq);
	return 0;
}

/* inlined */
static inline int kgr_btrfs_next_item(struct btrfs_root *root, struct btrfs_path *p)
{
	return kgr_btrfs_next_old_item(root, p, 0);
}

/* inlined */
static int kgr_qgroup_fix_relocated_data_extents(struct btrfs_trans_handle *trans,
					     struct reloc_control *rc)
{
	struct btrfs_fs_info *fs_info = rc->extent_root->fs_info;
	struct inode *inode = rc->data_inode;
	struct btrfs_root *data_reloc_root = BTRFS_I(inode)->root;
	struct btrfs_path *path;
	struct btrfs_key key;
	int ret = 0;

	if (!fs_info->quota_enabled)
		return 0;

	/*
	 * Only for stage where we update data pointers the qgroup fix is
	 * valid.
	 * For MOVING_DATA stage, we will miss the timing of swapping tree
	 * blocks, and won't fix it.
	 */
	if (!(rc->stage == UPDATE_DATA_PTRS && rc->extents_found))
		return 0;

	path = kgr_btrfs_alloc_path();
	if (!path)
		return -ENOMEM;
	key.objectid = kgr_btrfs_ino(inode);
	key.type = BTRFS_EXTENT_DATA_KEY;
	key.offset = 0;

	ret = kgr_btrfs_search_slot(NULL, data_reloc_root, &key, path, 0, 0);
	if (ret < 0)
		goto out;

	kgr_lock_extent(&BTRFS_I(inode)->io_tree, 0, (u64)-1);
	while (1) {
		struct btrfs_file_extent_item *fi;

		kgr_btrfs_item_key_to_cpu(path->nodes[0], &key, path->slots[0]);
		if (key.objectid > kgr_btrfs_ino(inode))
			break;
		if (key.type != BTRFS_EXTENT_DATA_KEY)
			goto next;
		fi = btrfs_item_ptr(path->nodes[0], path->slots[0],
				    struct btrfs_file_extent_item);
		if (btrfs_file_extent_type(path->nodes[0], fi) !=
				BTRFS_FILE_EXTENT_REG)
			goto next;
		/*
		pr_info("disk bytenr: %llu, num_bytes: %llu\n",
			btrfs_file_extent_disk_bytenr(path->nodes[0], fi),
			btrfs_file_extent_disk_num_bytes(path->nodes[0], fi));
			*/
		ret = kgr_btrfs_qgroup_insert_dirty_extent(trans, fs_info,
			btrfs_file_extent_disk_bytenr(path->nodes[0], fi),
			btrfs_file_extent_disk_num_bytes(path->nodes[0], fi),
			GFP_NOFS);
		if (ret < 0)
			break;
next:
		ret = kgr_btrfs_next_item(data_reloc_root, path);
		if (ret < 0)
			break;
		if (ret > 0) {
			ret = 0;
			break;
		}
	}
	kgr_unlock_extent(&BTRFS_I(inode)->io_tree, 0 , (u64)-1);
out:
	kgr_btrfs_free_path(path);
	return ret;
}

/* patched function */
int kgr_relocate_block_group(struct reloc_control *rc)
{
	struct rb_root blocks = RB_ROOT;
	struct btrfs_key key;
	struct btrfs_trans_handle *trans = NULL;
	struct btrfs_path *path;
	struct btrfs_extent_item *ei;
	u64 flags;
	u32 item_size;
	int ret;
	int err = 0;
	int progress = 0;

	path = kgr_btrfs_alloc_path();
	if (!path)
		return -ENOMEM;
	path->reada = 1;

	ret = kgr_prepare_to_relocate(rc);
	if (ret) {
		err = ret;
		goto out_free;
	}

	while (1) {
		rc->reserved_bytes = 0;
		ret = kgr_btrfs_block_rsv_refill(rc->extent_root,
					rc->block_rsv, rc->block_rsv->size,
					BTRFS_RESERVE_FLUSH_ALL);
		if (ret) {
			err = ret;
			break;
		}
		progress++;
		trans = kgr_btrfs_start_transaction(rc->extent_root, 0);
		if (IS_ERR(trans)) {
			err = PTR_ERR(trans);
			trans = NULL;
			break;
		}
restart:
		if (kgr_update_backref_cache(trans, &rc->backref_cache)) {
			kgr_btrfs_end_transaction(trans, rc->extent_root);
			continue;
		}

		ret = kgr_find_next_extent(rc, path, &key);
		if (ret < 0)
			err = ret;
		if (ret != 0)
			break;

		rc->extents_found++;

		ei = btrfs_item_ptr(path->nodes[0], path->slots[0],
				    struct btrfs_extent_item);
		item_size = btrfs_item_size_nr(path->nodes[0], path->slots[0]);
		if (item_size >= sizeof(*ei)) {
			flags = btrfs_extent_flags(path->nodes[0], ei);
			ret = kgr_check_extent_flags(flags);
			BUG_ON(ret);

		} else {
#ifdef BTRFS_COMPAT_EXTENT_TREE_V0
			u64 ref_owner;
			int path_change = 0;

			BUG_ON(item_size !=
			       sizeof(struct btrfs_extent_item_v0));
			ret = kgr_get_ref_objectid_v0(rc, path, &key, &ref_owner,
						  &path_change);
			if (ret < 0) {
				err = ret;
				break;
			}
			if (ref_owner < BTRFS_FIRST_FREE_OBJECTID)
				flags = BTRFS_EXTENT_FLAG_TREE_BLOCK;
			else
				flags = BTRFS_EXTENT_FLAG_DATA;

			if (path_change) {
				kgr_btrfs_release_path(path);

				path->search_commit_root = 1;
				path->skip_locking = 1;
				ret = kgr_btrfs_search_slot(NULL, rc->extent_root,
							&key, path, 0, 0);
				if (ret < 0) {
					err = ret;
					break;
				}
				BUG_ON(ret > 0);
			}
#else
			BUG();
#endif
		}

		if (flags & BTRFS_EXTENT_FLAG_TREE_BLOCK) {
			ret = kgr_add_tree_block(rc, &key, path, &blocks);
		} else if (rc->stage == UPDATE_DATA_PTRS &&
			   (flags & BTRFS_EXTENT_FLAG_DATA)) {
			ret = kgr_add_data_references(rc, &key, path, &blocks);
		} else {
			kgr_btrfs_release_path(path);
			ret = 0;
		}
		if (ret < 0) {
			err = ret;
			break;
		}

		if (!RB_EMPTY_ROOT(&blocks)) {
			ret = kgr_relocate_tree_blocks(trans, rc, &blocks);
			if (ret < 0) {
				/*
				 * if we fail to relocate tree blocks, force to update
				 * backref cache when committing transaction.
				 */
				rc->backref_cache.last_trans = trans->transid - 1;

				if (ret != -EAGAIN) {
					err = ret;
					break;
				}
				rc->extents_found--;
				rc->search_start = key.objectid;
			}
		}

		kgr_btrfs_end_transaction_throttle(trans, rc->extent_root);
		kgr_btrfs_btree_balance_dirty(rc->extent_root);
		trans = NULL;

		if (rc->stage == MOVE_DATA_EXTENTS &&
		    (flags & BTRFS_EXTENT_FLAG_DATA)) {
			rc->found_file_extent = 1;
			ret = kgr_relocate_data_extent(rc->data_inode,
						   &key, &rc->cluster);
			if (ret < 0) {
				err = ret;
				break;
			}
		}
	}
	if (trans && progress && err == -ENOSPC) {
		ret = kgr_btrfs_force_chunk_alloc(trans, rc->extent_root,
					      rc->block_group->flags);
		if (ret == 1) {
			err = 0;
			progress = 0;
			goto restart;
		}
	}

	kgr_btrfs_release_path(path);
	kgr_clear_extent_bits(&rc->processed_blocks, 0, (u64)-1, EXTENT_DIRTY,
			  GFP_NOFS);

	if (trans) {
		kgr_btrfs_end_transaction_throttle(trans, rc->extent_root);
		kgr_btrfs_btree_balance_dirty(rc->extent_root);
	}

	if (!err) {
		ret = kgr_relocate_file_extent_cluster(rc->data_inode,
						   &rc->cluster);
		if (ret < 0)
			err = ret;
	}

	rc->create_reloc_tree = 0;
	kgr_set_reloc_control(rc);

	kgr_backref_cache_cleanup(&rc->backref_cache);
	kgr_btrfs_block_rsv_release(rc->extent_root, rc->block_rsv, (u64)-1);

	err = kgr_prepare_to_merge(rc, err);

	kgr_merge_reloc_roots(rc);

	rc->merge_reloc_tree = 0;
	kgr_unset_reloc_control(rc);
	kgr_btrfs_block_rsv_release(rc->extent_root, rc->block_rsv, (u64)-1);

	/* get rid of pinned extents */
	trans = kgr_btrfs_join_transaction(rc->extent_root);
	if (IS_ERR(trans)) {
		err = PTR_ERR(trans);
		goto out_free;
	}
	/*
	 * Fix bsc#1008284
	 *  -err = kgr_qgroup_fix_relocated_data_extents(trans, rc);
	 *  -if (err < 0) {
	 *  -	btrfs_abort_transaction(trans, rc->extent_root, err);
	 *  +5 lines
	 */
	ret = kgr_qgroup_fix_relocated_data_extents(trans, rc);
	if (ret < 0) {
		btrfs_abort_transaction(trans, rc->extent_root, ret);
		if (!err)
			err = ret;
		goto out_free;
	}
	kgr_btrfs_commit_transaction(trans, rc->extent_root);
out_free:
	kgr_btrfs_free_block_rsv(rc->extent_root, rc->block_rsv);
	kgr_btrfs_free_path(path);
	return err;
}

static int kgr_patch_bsc1008284_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1008284_module_notify(struct notifier_block *nb,
        unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1008284_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1008284_module_nb = {
	.notifier_call = kgr_patch_bsc1008284_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1008284_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1008284_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1008284_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1008284_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1008284_module_nb);
}
